#! /usr/bin/env python
# -*- coding: utf-8 -*-

# logging
import logging

# requirements
from pandas import DataFrame

# part of module
from .counts import score_counts

logger = logging.getLogger(__name__)


# TODO re-write as regular methods
class Keywords:
    """ keyword analysis """

    def __init__(self, corpus, df_dump, p_query=['lemma']):

        self.corpus = corpus

        # activate dump
        self.size = len(df_dump)

        # consistency check
        if self.size == 0:
            logger.warning('cannot calculate keywords on 0 regions')
            return

        # determine layer to work on
        self.p_query = [p_query] if isinstance(p_query, str) else p_query
        available_attributes = self.corpus.attributes_available['attribute'].values
        if not set(self.p_query).issubset(set(available_attributes)):
            logger.warning('specfied p-attribute(s) ("%s") not available' % " ".join(self.p_query))
            logger.warning('falling back to primary layer')
            self.p_query = ['word']

        # collect context and save result
        logger.info('collecting token counts of subcorpus')
        self.counts = corpus.counts.dump(
            df_dump, start='match', end='matchend', p_atts=self.p_query, split=True
        )

    def show(self, order='log_likelihood', cut_off=100, ams=None,
             min_freq=2, frequencies=True, flags=None,
             marginals='corpus', show_negative=False):

        # consistency check
        if self.counts.empty:
            logger.warning("nothing to show")
            return DataFrame()

        # get subcorpus frequencies und apply min freq
        f = self.counts.rename(columns={'freq': 'f'})
        vocab = len(f)
        f1 = f['f'].sum()
        f = f.loc[f['f'] >= min_freq]

        # get reference frequency
        if isinstance(marginals, str):
            if marginals == 'corpus':
                N = self.corpus.corpus_size
                marginals = self.corpus.marginals(f.index, self.p_query)
            else:
                raise NotImplementedError
        elif isinstance(marginals, DataFrame):
            # DataFrame must contain a column 'freq'
            N = marginals['freq'].sum()
        else:
            raise NotImplementedError

        # create dataframe
        f2 = marginals[['freq']].rename(columns={'freq': 'f2'})
        df = f2.join(f)
        df = df.fillna(0)
        df['f1'] = f1
        df['N'] = N

        # score
        keywords = score_counts(df, order=order, cut_off=cut_off,
                                flags=flags, ams=ams, vocab=vocab)

        if frequencies:
            # throw away anti-keywords by default
            if not show_negative:
                keywords = keywords.loc[keywords['O11'] >= keywords['E11']]

        return keywords


def keywords(corpus, corpus_reference, p, p_reference, order='O11',
             cut_off=100, ams=None, min_freq=2, frequencies=True, flags=None):
    """directly extract keywords by comparing two (sub-)corpora

    :param ccc.Corpus corpus: target corpus
    :param ccc.Corpus corpus_reference: reference corpus
    :param str p: p-att
    :param str p_reference: p-att
    :param str order: AM to order
    :param list ams: list of AMs
    :param int min_freq: minimum number of occurrences in target corpus
    :param bool frequencies: add frequency columns
    :param list flags: '%cd'

    """

    # get and merge both dataframes of counts
    left = corpus.marginals(p_atts=p)[['freq']].rename(columns={'freq': 'f1'})
    right = corpus_reference.marginals(p_atts=p_reference)[['freq']].rename(columns={'freq': 'f2'})
    df = left.join(right, how='outer')
    df['N1'] = left['f1'].sum()
    df['N2'] = right['f2'].sum()
    df = df.fillna(0)

    # score counts
    keywords = score_counts(df, order=order, cut_off=cut_off,
                            flags=flags, ams=ams, digits=4)

    return keywords



#! /usr/bin/env python
# -*- coding: utf-8 -*-
"""
__init__.py is a special Python file that allows a directory to become
a Python package so it can be accessed using the 'import' statement.
"""

from .cwb import Corpora, Corpus
from .version import __version__



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import os
import shelve
from glob import glob
from hashlib import sha256

logger = logging.getLogger(__name__)


def generate_idx(identifiers, prefix='', length=10):

    string = ''.join([str(idx) for idx in identifiers])
    identifier = sha256(str(string).encode()).hexdigest()
    return prefix + identifier[:length]


def generate_library_idx(lib_path, prefix='', length=10):

    wordlists = glob(os.path.join(lib_path, 'wordlists', '*.txt'))
    macros = glob(os.path.join(lib_path, 'macros', '*.txt'))
    paths = macros + wordlists
    identifiers = [
        generate_idx(open(p, 'rt')) for p in sorted(paths)
    ]
    return generate_idx(identifiers, prefix, length)


class Cache:

    def __init__(self, path=None):

        self.path = path

        if path:
            directory = os.path.dirname(path)
            os.makedirs(directory, exist_ok=True)

    def delete(self, identifier):

        if self.path is None:
            logger.info('no cache path')
            return

        if isinstance(identifier, str):
            key = identifier
        else:
            key = generate_idx(identifier)

        with shelve.open(self.path) as shelf:
            if key in shelf.keys():
                logger.info('deleting object "%s" from cache' % key)
                del shelf[key]

    def get(self, identifier):

        if self.path is None:
            logger.info('no cache path')
            return

        if isinstance(identifier, str):
            key = identifier
        else:
            key = generate_idx(identifier)

        with shelve.open(self.path) as shelf:
            if key in shelf.keys():
                logger.info('retrieving object "%s" from cache' % key)
                return shelf[key]
            else:
                return None

    def set(self, identifier, value):

        if self.path is None:
            logger.info('no cache path')
            return

        if isinstance(identifier, str):
            key = identifier
        else:
            key = generate_idx(identifier)

        with shelve.open(self.path) as shelf:
            logger.info('saving object "%s" to cache' % key)
            shelf[key] = value



#! /usr/bin/env python
# -*- coding: utf-8 -*-

""" frames.py: classes for dataframe types

### DumpFrame ###

=== CREATION ===

    dump_from_query
    === (match, matchend), 0*, .., 9* ===
    index <int, int> match, matchend: cpos
    columns <int> 0*, .., 9*: anchor points (optional; missing = -1)

    dump_from_s_att
    === (match, matchend), $s_cwbid, $s* ===
    index <int, int> match, matchend: cpos
    column <int> $s_cwbid: id
    column <str> $s: annotation (optional)
    no missing values

=== TRANSFORMATION (additional columns are preserved) ===

    dump2patt
    === (match, matchend), $p ===
    index <int, int> match, matchend: cpos
    column <str> $p: " "-joined tokens (match..matchend or regions of input columns)

    dump2satt
    === (match, matchend), $s_cwbid, $s_span, $s_spanend, $s* ===
    index <int, int> match, matchend: cpos
    column <int> $s_cwbid: id (missing = -1)
    column <int> $s_span: cpos (missing = -1)
    column <int> $s_spanend: cpos (missing = -1)
    column <str> $s*: annotation (optional)

    dump2context
    === (match, matchend), contextid*, context, contextend
        $s_cwbid*, $s_span*, $s_spanend* ===
    index <int, int> match, matchend: cpos
    column <int> contextid: of id (optional; duplicate of $s_cwbid)
    column <int> context: cpos
    column <int> contextend: cpos
    column <int> $s_cwbid: id (optional; missing = -1)
    column <int> $s_span: cpos (missing = -1)
    column <int> $s_spanend: cpos (missing = -1)

=== QUERY ===

    query
    === (match, matchend), 0*, ..., 9*, contextid*, context*, contextend* ===
    index <int, int> match, matchend: cpos
    columns <int> 0*, .., 9*: anchor points (optional; missing = -1)
    column <int> contextid: of id (optional)
    column <int> context: cpos (optional)
    column <int> contextend: cpos (optional)

    query_s_att
    === (match, matchend), $s_cwbid, $s* ===
    index <int, int> match, matchend: cpos
    column <int> $s_cwbid: id
    column <str> $s: annotation (optional)
    no missing values


### ConcFrame ###

    __init__ enforces context, contextend columns (optional = match, matchend)

    === (match, matchend) context contextend ===

    simple
    === (match, matchend) $p ===

    kwic
    === (match, matchend) $p_left, $p_node, $p_right ===

    slots
    === (match, matchend) $p_$slot[0]..$slot[1] ===

    dict
    === (match, matchend) dict ===

    dataframe
    === (match, matchend) dataframe ===

### FreqFrame ###

    returned by
    - Counts.cpos()
    - Counts.dump()
      + 1 counts._cpos2patts + counts.count_items
      + 2 write dump to tempfile + counts.cwb_scan_corpus
    - Counts.matches()
      + 1 cqp.count
      + 2 cqp.tabulate + counts.count_items
      + 3 cqp.dump + counts.cwb_scan_corpus
    - Counts.mwus()
      + 1 for each query: cqp.size
      + 2 for "|".joined-query: counts.dump
      + 3 for "|".joined-query: counts.matches
    - Corpus.marginals()

    == (item) freq, p_atts[0]*, p_atts[1]*, ... ==


### ScoreFrame ###

    returned by score_counts()

    == (item) freq, O11*..O22*, E11*..E22*, AM1, AM2, ... ==



### DumpsFrame ###

=== CREATION ===

    topic (= dump)
    === (match, matchend) contextid context contextend ===
    note that the contexts may overlap and even include other matches

    complex node via cluster id
    === (match, matchend) context contextend contextid match matchend offset ===

    second discourseme (= dump)
    === (match1, match1end) contextid ===

    here: discourseme co-occurs with topic, iff discourseme in context(topic)
    here: context(topic) for second-order collocation analysis
    other possibilities: context(topic & discourseme), c(t | d), ...
    left_join via contextid

    === (match, matchend) context contextend contextid match1 match1end offset1 ===

    normal row
    == (m, me) c ce ci m1 m1e o1 ==
       (5,  6) 0 10  1  2   2 -3
    surface realization of topic on cpos 5-6, discourseme 1 on cpos 2

    duplicate rows for several co-occurrences, e.g.
    == (m, me) c ce ci m1 m1e o1 ==
       (5,  6) 0 10  1  2   2 -3
       (5,  6) 0 10  1  8   8 +2
    another occurrence of discourseme 1 on cpos 8

    slicing further discoursemes:

    === (m, me) c ce ci m1 m1e m2 m2e m3 m3e ===

        (5,  6) 0 10  1  2   2 -3 3 3 -2 0 0 -5
        (5,  6) 0 10  1  8   8 +2 3 3 -2 0 0 -5
        (5,  6) 0 10  1  2   2 -3 7 7 +2 0 0 -5
        (5,  6) 0 10  1  8   8  +2 7 7 +2 0 0 -5


### CoocFrame ###





"""



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import logging

# requirements
from association_measures.measures import score
from pandas import NA, DataFrame, concat

# part of module
from . import Corpus
from .collocates import Collocates, dump2cooc
from .concordances import Concordance
from .dumps import Dump
from .utils import format_cqp_query

logger = logging.getLogger(__name__)


########################################################
# (1) FOCUS ON TOPIC: DIRECTED (IE INDEXED) DATAFRAMES #
########################################################

# init(dump, name='topic')
# input : dump.df: === (m, me) ci c ce ==
# output: self.df: === (m, me) ci c ce m_topic me_topic o_topic ===
# with duplicate indices:

#   (m   me)     ci        c       ce      m_t   me_t   o_t
#  638  638      18      629      672      638    638     0
#       638      18      629      672      640    640     2
#  640  640      18      629      672      638    638    -2
#       640      18      629      672      640    640     0
# 1202 1202      36     1195     1218     1202   1202     0
# ...   ...     ...      ...      ...      ...    ...   ...

# #### 1 occurrence
# ... die [CDU]:m-me+m_t-me_t will eigentlich ...

# #### 2 occurrences
# ... die [CDU]:m-me+m_t-me_t und [CSU] gehen da ...  0
# ... die [CDU]:m-me und [CSU]:m_t-me_t gehen da ...  2
# ... die [CDU]:m_t-me_t und [CSU]:m-me gehen da ... -2
# ... die [CDU] und [CSU]:m-me+m_t-me_t gehen da ...  0

# #### 3 occurrences
# ... die [Union]:0+1, d.h. die [CDU] und [CSU] gehen da ...
# ... die [Union]:0, d.h. die [CDU]:1 und [CSU] gehen da ...
# ... die [Union]:0, d.h. die [CDU] und [CSU]:1 gehen da ...
# ... die [Union]:1, d.h. die [CDU]:0 und [CSU] gehen da ...
# ... die [Union], d.h. die [CDU]:0+1 und [CSU] gehen da ...
# ... die [Union], d.h. die [CDU]:0 und [CSU]:1 gehen da ...
# ... die [Union]:1, d.h. die [CDU] und [CSU]:0 gehen da ...
# ... die [Union], d.h. die [CDU]:1 und [CSU]:0 gehen da ...
# ... die [Union], d.h. die [CDU] und [CSU]:0+1 gehen da ...

# add_discourseme(dump, name, drop=True)
# input:  === (m, me) ci c ce m_t me_t o_t ==
#         === (m_d, me_d) ci ==
# output: === (m, me) ci c ce m_t me_t o_t m_d me_d o_d ==

#   (m   me)     ci       c      ce    m_t   me_t    o_t     m_d   me_d   o_d
#  638  638      18     629     672    638    638      0     636    636    -2
#       638      18     629     672    640    640      2     636    636    -2
#  640  640      18     629     672    638    638     -2     636    636    -4
#       640      18     629     672    640    640      0     636    636    -4
# 1202 1202      36    1195    1218   1202   1202      0    1205   1205     3
#  ...  ...     ...     ...     ...    ...    ...    ...     ...    ...   ...

# #### t . d
# #### d . t
# ... die [CDU]:0+1 [will]:2 eigentlich ...

# #### t . t . d
# #### t . d . t
# #### d . t . t
# ... die [CDU]:0+1 und [CSU] [will]:2 eigentlich ...
# ... die [CDU]:0 und [CSU]:1 [will]:2 eigentlich ...
# ... die [CDU]:1 und [CSU]:0 [will]:2 eigentlich ...
# ... die [CDU] und [CSU]:0+1 [will]:2 eigentlich ...

# add_discourseme(dump, name, drop=False)

#     (m     me)     ci        c       ce     m_t    me_t      o_t    m_d    me_d    o_d
#    638    638      18      629      672     638     638        0    636     636     -2
#           638      18      629      672     640     640        2    636     636     -2
#    640    640      18      629      672     638     638       -2    636     636     -4
#           640      18      629      672     640     640        0    636     636     -4
#   1202   1202      36     1195     1218    1202    1202        0   1205    1205      3
# ...       ...     ...      ...      ...     ...     ...       ...   ...     ...    ...
# 149292 149292    7317   149286   149306  149290  149290       -2   <NA>    <NA>   <NA>
#        149292    7317   149286   149306  149292  149292        0   <NA>    <NA>   <NA>

# group_lines()
# === (m, me) ci c ce m0 m0e o0 m1 me1 o1 m2 me2 o2 ===
# with duplicate indices to
# === (m, me) ci c ce d0 d1 d2 ===
# without duplicate indices, where
# d0 = {(o0, m0, m0e), (o0, m0, m0e), ...}
# d1 = {(o1, m1, m1e), ...}
# ...

##########################################################
# (2) OUTER JOIN ON CONTEXTID: FULL TEXTUAL COOCCURRENCE #
##########################################################


# TODO discourseme definition: items, queries, corpora
class Discourseme:
    def __init__(self):
        pass


def constellation_left_join(df1, df2, name, drop=True, window=None):
    """join an additional dump df2 to an existing constellation

    :param DataFrame df1: constellation dump === (m, me) ci c ce m_t* me_t* o_t* m_1* me_1* o_1* ... ==
    :param DataFrame df2: additional dump === (m, me) ci c ce ==
    :param str name: name for additional discourseme
    :param bool drop: drop all rows that do not contain all discoursemes within topic context?
    :return: constellation dump including additional dump  === (m, me) ci c ce m_t me_t o_t m_1 me_1 o_1 ... m_name me_name o_name ==
    :rtype: DataFrame
    """

    # merge dumps via contextid ###
    df1 = df1.reset_index()
    df2 = df2.reset_index()[['contextid', 'match', 'matchend']].astype("Int64")
    m = df1.merge(df2, on='contextid', how='left')

    # calculate offset ###
    m['offset_y'] = 0       # init as overlap
    # y .. x
    m.loc[m['match_x'] > m['matchend_y'], 'offset_y'] = m['matchend_y'] - m['match_x']
    # x .. y
    m.loc[m['matchend_x'] < m['match_y'], 'offset_y'] = m['match_y'] - m['matchend_x']
    # missing y
    m.loc[m['match_y'].isna(), 'offset_y'] = NA

    # restrict to complete constellation ###
    if drop:
        m = m.dropna()
        if window is None:
            # only keep co-occurrences that are within context
            m = m.loc[
                (m['matchend_y'] >= m['context']) & (m['match_y'] < m['contextend'])
            ]
        else:
            m = m.loc[abs(m['offset_y']) <= window]

    # rename columns ###
    m = m.rename(columns={
        'match_x': 'match',
        'matchend_x': 'matchend',
        'match_y': 'match_' + name,
        'matchend_y': 'matchend_' + name,
        'offset_y': 'offset_' + name
    })

    # set index ###
    m = m.set_index(['match', 'matchend'])

    return m


def aggregate_matches(df, name, context_col='contextid',
                      match_cols=['match', 'matchend']):

    # counts
    counts = DataFrame(df[context_col].value_counts()).astype("Int64")
    counts.columns = ['COUNTS_' + name]

    # matches
    matches = df.reset_index()
    matches['MATCHES_' + name] = matches[match_cols].values.tolist()
    matches['MATCHES_' + name] = matches['MATCHES_' + name].apply(tuple)
    matches = matches.groupby('contextid')['MATCHES_' + name].apply(set)

    # combine
    table = counts.join(matches)

    return table


def format_roles(row, names, s_show, window, htmlify_meta=False):
    """Take a row of a dataframe indexed by match, matchend of the node,
    columns for each discourseme with sets of tuples indicating discourseme positions,
    columns for each s in s_show,
    and a column 'dict' containing the pre-formatted concordance line.

    creates a list (aligned with other lists) of lists of roles; roles are:
    - 'node' (if cpos in index)
    - 'out_of_window' (if offset of cpos from node > window)
    - discourseme names

    :return: concordance line for MMDA frontend
    :rtype: dict

    """

    # TODO directly create relevant objects, no need for frontend to take care of it
    # init
    d = row['dict']
    roles = list()

    # 'out_of_window' | None | 'node'
    role = ['out_of_window' if abs(t) > window else None for t in d['offset']]
    for i in range(d['cpos'].index(row.name[0]), d['cpos'].index(row.name[1]) + 1):
        role[i] = 'node'
    roles.append(role)

    # discourseme names
    for name in names:

        role = [None] * len(d['offset'])

        if not isinstance(row[name], float):
            for t in row[name]:

                # check match information
                if len(t) == 2:
                    # lazy definition without offset
                    start = 0
                    end = 1
                elif len(t) == 3:
                    # with offset
                    start = 1
                    end = 2
                else:
                    continue

                # skip NAs
                if not isinstance(t[start], int):
                    continue

                # skip the ones too far away
                try:
                    start = d['cpos'].index(t[start])
                    end = d['cpos'].index(t[end]) + 1
                except ValueError:
                    continue

                for i in range(start, end):
                    role[i] = name

        roles.append(role)

    # combine individual roles into one list of lists
    d['role'] = [[a for a in set(r) if a is not None] for r in list(zip(*roles))]

    # add s-attributes
    if htmlify_meta:
        meta = {key: row[key] for key in s_show if not key.startswith("BOOL")}
        d['meta'] = DataFrame.from_dict(
            meta, orient='index'
        ).to_html(bold_rows=False, header=False)
        for s in s_show:
            if s.startswith("BOOL"):
                d[s] = row[s]
    else:
        for s in s_show:
            d[s] = row[s]

    return d


class Constellation:

    def __init__(self, dump, name='topic'):
        """
        param Dump dump: dump with dump.corpus, dump.df: == (m, me) ci, c, ce ==
        param str name: name of the node
        """

        self.corpus = dump.corpus
        # init with given dump
        self.df = dump.df[['contextid', 'context', 'contextend']].astype("Int64")
        # init added discoursemes
        self.discoursemes = {}
        # the topic is treated as common discourseme
        self.add_discourseme(dump, name=name)

    def __str__(self):
        return (
            "\n" + "a constellation with %d match indices" % len(self.df) + "\n" +
            "%d registered discourseme(s) " % len(self.discoursemes) + "\n" +
            "\n".join(["- '%s' with %d matches" % (d, len(self.discoursemes[d].df))
                       for d in self.discoursemes])
        )

    def __repr__(self):
        """Info string

        """
        return self.__str__()

    def add_discourseme(self, dump, name='discourseme', drop=True, window=None):
        """
        :param Dump dump: dump.df: == (m, me) ci ==
        :param str name: name of the discourseme
        :param bool drop: remove matches without all discoursemes in node context
        """

        # register discourseme
        if name in self.discoursemes.keys():
            logger.error('name "%s" already taken; cannot register discourseme' % name)
            return

        self.discoursemes[name] = dump
        self.df = constellation_left_join(self.df, dump.df, name,
                                          drop=drop, window=window)

    def group_lines(self):
        """
        convert dataframe:
        === (m, me) ci c ce m0 m0e o0 m1 me1 o1 m2 me2 o2 ===
        with duplicate indices to
        === (m, me) ci c ce m0 m1 m2 ===
        without duplicate indices
        where
        m0 = {(o0, m0, m0e), (o0, m0, m0e), ...}
        m1 = {(o1, m1, m1e), ...}

        """

        df = self.df.copy()
        df_reduced = df[~df.index.duplicated(keep='first')][
            ['contextid', 'context', 'contextend']
        ]
        for name in self.discoursemes.keys():
            columns = [m + "_" + name for m in ['offset', 'match', 'matchend']]
            df[name] = df[columns].values.tolist()
            df[name] = df[name].apply(tuple)
            df = df.drop(columns, axis=1)
            df_reduced[name] = df.groupby(level=['match', 'matchend'])[name].apply(set)

        return df_reduced

    def concordance(self, window=5,
                    p_show=['word', 'lemma'], s_show=[],
                    order='random', cut_off=100, random_seed=42,
                    htmlify_meta=False):
        """Retrieve concordance lines for constellation.

        :param int window: cpos further away from node will be marked 'out_of_window'

        :return: concordance lines
        :rtype: list
        """

        # convert dataframe
        df = self.group_lines()

        # cut off and sampling (done here to be able to use random_seed)
        cut_off = len(df) if cut_off is None or cut_off > len(df) else cut_off
        if order == 'random':
            df = df.sample(cut_off, random_state=random_seed)
        elif order == 'first':
            df = df.head(cut_off)
        elif order == 'last':
            df = df.head(cut_off)
        else:
            raise NotImplementedError

        # retrieve concordance lines
        conc = Concordance(self.corpus.copy(), df)
        lines = conc.lines(form='dict', p_show=p_show, s_show=s_show,
                           order=order, cut_off=cut_off)

        # map roles
        output = lines.apply(
            lambda row: format_roles(
                row, self.discoursemes.keys(), s_show, window, htmlify_meta
            ), axis=1
        )

        return list(output)

    def collocates(self, windows=[3, 5, 7],
                   p_show=['lemma'], flags=None,
                   ams=None, frequencies=True,
                   min_freq=2, order='log_likelihood', cut_off=None,
                   marginals='corpus'):
        """Retrieve collocates for constellation.

        :param int window: window around node for pre-selected matches

        :return: collocates
        :rtype: list of DataFrames
        """

        # get relevant contexts
        df_dump = self.df.drop_duplicates(subset=['context', 'contextend'])
        df_cooc, f1_set = dump2cooc(df_dump)

        logger.info('get cpos that are consumed by discoursemes')
        for idx in self.discoursemes.keys():
            f1_set.update(self.discoursemes[idx].matches())

        # correct df_cooc
        df_cooc = df_cooc.loc[~df_cooc['cpos'].isin(f1_set)]

        # count once
        node_freq = self.corpus.counts.cpos(f1_set, p_show)

        collocates = Collocates(
            corpus=self.corpus, df_dump=None, p_query=p_show, mws=max(windows),
            df_cooc=df_cooc, f1_set=f1_set, node_freq=node_freq
        )

        output = dict()
        for window in windows:
            output[window] = collocates.show(
                window=window, order=order, cut_off=cut_off, ams=ams,
                min_freq=min_freq, frequencies=frequencies, flags=flags,
                marginals=marginals
            )

        return output


class TextConstellation:

    def __init__(self, dump, s_context, name='topic'):
        """
        param Dump dump: dump with dump.corpus, dump.df: == (m, me) ci, c, ce ==
        param str name: name of the node
        """

        if isinstance(dump, dict):
            dumps = dump.copy()
            name = list(dumps.keys())[0]
            dump = dumps.pop(name)

        elif isinstance(dump, Dump):
            dumps = {}

        else:
            raise ValueError()

        # create
        self.s_context = s_context
        self.corpus = dump.corpus
        self.N = len(self.corpus.attributes.attribute(s_context, 's'))

        try:
            self.df = aggregate_matches(dump.df, name)
        except KeyError:        # no matches
            self.df = DataFrame()

        for name, dump in dumps.items():
            self.add_discourseme(dump, name)

    def add_discourseme(self, dump, name='discourseme'):

        # register discourseme
        if name in self.df.columns:
            logger.error('name "%s" already taken; cannot register discourseme' % name)
            return

        try:
            dump = dump.set_context(context_break=self.s_context)
            df = aggregate_matches(dump.df, name)
        except KeyError:        # no matches
            df = DataFrame()

        if df.empty:
            return
        elif self.df.empty:
            self.df = df
        else:
            df = self.df.join(df, how='outer')

        self.df = df.sort_index()

    def breakdown(self, p_atts=['word']):

        print(self.df.columns)
        raise NotImplementedError()

    def concordance(self, window=0,
                    p_show=['word', 'lemma'], s_show=[],
                    order='random', cut_off=100, random_seed=42,
                    htmlify_meta=False, form='list'):

        # cut off and sampling
        cut_off = len(self.df) if cut_off is None or cut_off > len(self.df) else cut_off
        if order == 'random':
            df = self.df.sample(cut_off, random_state=random_seed)
        elif order == 'first':
            df = self.df.head(cut_off)
        elif order == 'last':
            df = self.df.head(cut_off)
        else:
            raise NotImplementedError

        # join context..contextend
        contexts = self.corpus.dump_from_s_att(self.s_context, annotation=False)
        contexts.columns = ['contextid']
        contexts = contexts.reset_index().set_index('contextid')
        df = df.join(contexts).set_index(['match', 'matchend'])

        # retrieve concordance lines
        conc = Concordance(self.corpus.copy(), df)
        lines = conc.lines(form='dict', p_show=p_show, s_show=s_show,
                           order=order, cut_off=cut_off)

        # get boolean columns for each discourseme
        names_bool = list()
        for name in [c for c in df.columns if c.startswith("COUNTS_")]:
            name_bool = '_'.join(['BOOL', name.split("COUNTS_")[-1]])
            names_bool.append(name_bool)
            lines[name_bool] = (lines[name] > 0)
            lines[name_bool] = lines[name_bool].fillna(False)

        # format roles
        match_cols = [c for c in lines.columns if c.startswith("MATCHES_")]
        match_names = [c.split("MATCHES_")[-1] for c in match_cols]
        col_mapper = dict(zip(match_cols, match_names))
        lines = lines.rename(columns=col_mapper)

        if form == 'table':
            return lines
        elif form == 'list':
            output = lines.apply(
                lambda row: format_roles(
                    row, match_names, s_show=names_bool+s_show, window=window, htmlify_meta=htmlify_meta
                ), axis=1
            )
            return list(output)
        else:
            raise NotImplementedError()

    def associations(self, ams=None, frequencies=True,
                     min_freq=2, order='log_likelihood',
                     cut_off=None):

        counts = self.df[[c for c in self.df.columns if c.startswith("COUNTS_")]]
        counts.columns = [c.split("COUNTS_")[-1] for c in counts.columns]
        cooc = counts > 0

        # TODO triangular matrix
        tables = DataFrame()
        for name in counts.columns:
            table = round(textual_associations(
                cooc, self.N, name
            ).reset_index(), 2)
            table['node'] = name
            tables = concat([tables, table])

        tables = tables[
            ['node', 'candidate'] +
            [d for d in tables.columns if d not in ['node', 'candidate']]
        ]

        return tables


def textual_associations(cooc, N, column):
    """Textual association.

    """
    f1 = cooc[column].sum()
    candidates = [c for c in cooc.columns if c != column]
    records = list()
    for candidate in candidates:
        f2 = cooc[candidate].sum()
        f = (cooc[[column, candidate]].sum(axis=1) == 2).sum()
        records.append({
            'candidate': candidate,
            'f1': f1,
            'f2': f2,
            'f': f,
            'N': N
        })
    contingencies = DataFrame(records).set_index('candidate')
    measures = score(contingencies, freq=True)
    contingencies = contingencies.join(measures.drop(['N'], axis=1))

    return contingencies


########################################################
# CONSTELLATION CREATION FROM DISCOURSEMES AND QUERIES
# TODO parallelize
########################################################
def create_constellation(corpus_name,
                         # discoursemes
                         topic_discourseme,
                         filter_discoursemes,
                         additional_discoursemes,
                         # context settings
                         s_context,
                         context=20,
                         # query settings
                         p_query='word',
                         s_query=None,
                         flags='%cd',
                         escape=True,
                         match_strategy='longest',
                         # CWB settings
                         lib_path=None,
                         cqp_bin='cqp',
                         registry_path='/usr/local/share/cwb/registry/',
                         data_path='/tmp/ccc-data/',
                         window=None,
                         approximate=False):
    """simple constellation creator. returns a Constellation() if a
    topic_discourseme is given, otherwise a TextConstellation(). Note
    that for TextConstellations, there is no difference between
    additional discoursemes and filter discoursemes (instead, a
    boolean column for each discourseme is added).

    :param dict topic_discourseme: used for init
    :param dict filter_discoursemes: inner join
    :param dict additional_discourseme: left join

    """

    # pre-process parameters
    s_context = s_query if not s_context else s_context
    s_query = s_context if s_query is None else s_query

    # init corpus
    corpus = Corpus(corpus_name, lib_path, cqp_bin, registry_path, data_path)

    # topic -> Constellation()
    if len(topic_discourseme) > 0:

        if len(topic_discourseme) > 1:
            raise ValueError("only one topic discourseme can be given")

        # init with topic
        topic_name = list(topic_discourseme.keys())[0]
        topic_items = topic_discourseme[topic_name]
        topic_query = format_cqp_query(
            topic_items, p_query=p_query, s_query=s_query, flags=flags, escape=escape
        )
        topic_dump = corpus.query(
            topic_query,
            context=context,
            context_break=s_context,
            match_strategy=match_strategy
        )
        const = Constellation(topic_dump, topic_name)

        if approximate:
            sub = topic_dump.df.set_index(['context', 'contextend'])
            corpus.activate_subcorpus(nqr='TempRestriction', df_dump=sub)
        # add filter discoursemes
        for disc_name, disc_items in filter_discoursemes.items():
            disc_query = format_cqp_query(
                disc_items, p_query=p_query, s_query=s_query, flags=flags, escape=escape
            )
            disc_dump = corpus.query(
                disc_query,
                context=None,
                context_break=s_context,
                match_strategy=match_strategy
            )
            if len(disc_dump.df) > 0:
                const.add_discourseme(disc_dump, disc_name, drop=True, window=window)
            else:
                raise ValueError()

        # add additional discoursemes
        for disc_name, disc_items in additional_discoursemes.items():
            disc_query = format_cqp_query(
                disc_items, p_query=p_query, s_query=s_query, flags=flags, escape=escape
            )
            disc_dump = corpus.query(
                disc_query,
                context=None,
                context_break=s_context,
                match_strategy=match_strategy
            )
            if len(disc_dump.df) > 0:
                const.add_discourseme(disc_dump, disc_name, drop=False)
        corpus.activate_subcorpus()

    # no topic -> TextConstellation()
    else:

        # no filter implemented: all discoursemes are equal
        discoursemes = {**filter_discoursemes, **additional_discoursemes}

        # init with arbitrary topic
        topic_name = list(discoursemes.keys())[0]
        topic_items = discoursemes.pop(topic_name)
        topic_query = format_cqp_query(
            topic_items, p_query=p_query, s_query=s_query, flags=flags, escape=escape
        )
        topic_dump = corpus.query(
            topic_query,
            context=context,
            context_break=s_context,
            match_strategy=match_strategy
        )
        const = TextConstellation(topic_dump, s_context, topic_name)

        # add further discoursemes
        for disc_name, disc_items in discoursemes.items():
            disc_query = format_cqp_query(
                disc_items, p_query=p_query, s_query=s_query, flags=flags, escape=escape
            )
            disc_dump = corpus.query(
                disc_query,
                context=None,
                context_break=s_context,
                match_strategy=match_strategy
            )
            const.add_discourseme(disc_dump, disc_name)

    return const


def create_constellation_query(corpus_name,
                               # discoursemes
                               topic_discourseme,
                               filter_discoursemes,
                               additional_discoursemes,
                               # context settings
                               s_context,
                               context=20,
                               # CWB settings
                               match_strategy='longest',
                               lib_path=None,
                               cqp_bin='cqp',
                               registry_path='/usr/local/share/cwb/registry/',
                               data_path='/tmp/ccc-data/',
                               window=None):
    """same as above, but with pre-formatted CQP queries

    """

    # init corpus
    corpus = Corpus(corpus_name, lib_path, cqp_bin, registry_path, data_path)

    # topic -> Constellation()
    if len(topic_discourseme) > 0:

        if len(topic_discourseme) > 1:
            raise ValueError("only one topic discourseme can be given")

        # init with topic
        topic_name = list(topic_discourseme.keys())[0]
        topic_query = topic_discourseme[topic_name]
        topic_dump = corpus.query(
            topic_query,
            context=context,
            context_break=s_context,
            match_strategy=match_strategy
        )
        const = Constellation(topic_dump, topic_name)

        # add filter discoursemes
        for disc_name, disc_query in filter_discoursemes.items():
            disc_dump = corpus.query(
                disc_query,
                context=None,
                context_break=s_context,
                match_strategy=match_strategy
            )
            const.add_discourseme(disc_dump, disc_name, drop=True, window=window)

        # add additional discoursemes
        for disc_name, disc_query in additional_discoursemes.items():
            disc_dump = corpus.query(
                disc_query,
                context=None,
                context_break=s_context,
                match_strategy=match_strategy
            )
            const.add_discourseme(disc_dump, disc_name, drop=False)

    # no topic -> TextConstellation()
    else:

        # no filter implemented: all discoursemes are equal
        discoursemes = {**filter_discoursemes, **additional_discoursemes}

        # init with arbitrary topic
        topic_name = list(discoursemes.keys())[0]
        topic_query = discoursemes.pop(topic_name)
        topic_dump = corpus.query(
            topic_query,
            context=context,
            context_break=s_context,
            match_strategy=match_strategy
        )
        const = TextConstellation(topic_dump, s_context, topic_name)

        # add further discoursemes
        for disc_name, disc_query in discoursemes.items():
            disc_dump = corpus.query(
                disc_query,
                context=None,
                context_break=s_context,
                match_strategy=match_strategy
            )
            const.add_discourseme(disc_dump, disc_name)

    return const



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import logging

# requirements
from pandas import DataFrame

# part of module
from .collocates import Collocates
from .concordances import Concordance
from .keywords import Keywords
from .utils import correct_anchors, fold_df, merge_intervals

logger = logging.getLogger(__name__)


class Dump:
    """ result of a query """

    def __init__(self, corpus, df_dump, name_cqp):

        self.df = df_dump
        self.corpus = corpus
        self.name_cqp = name_cqp
        self.size = len(self.df)

        self._matches = None
        self._context = None

    def __str__(self):

        desc = ['a ccc.Dump with %d matches' % self.size]
        if self.corpus.subcorpus:
            crpssbcrps = self.corpus.corpus_name + ":" + self.corpus.subcorpus
        else:
            crpssbcrps = self.corpus.corpus_name
        desc.append('- corpus "%s" (%d tokens in whole corpus)' % (
            crpssbcrps, self.corpus.corpus_size
        ))
        if self.name_cqp is not None:
            desc.append('- name in cqp: "%s"' % self.name_cqp)
        desc.append('- columns: %s' % str(list(self.df.columns)))
        return "\n".join(desc)

    def __repr__(self):
        """Info string.

        """
        return self.__str__()

    def set_context(self, context=None, context_break=None,
                    context_left=None, context_right=None):
        """Set context in the dump.

        """
        # pre-process context
        if context_left is None:
            context_left = context
        if context_right is None:
            context_right = context

        # set context
        df = self.corpus.dump2context(
            self.df, context_left, context_right, context_break
        )

        return Dump(self.corpus, df, self.name_cqp)

    def correct_anchors(self, corrections):
        """Correct anchors by integer offsets.

        """
        self.df = correct_anchors(self.df, corrections)

    def breakdown(self, p_atts=['word'], flags=""):
        """Frequency breakdown of match..matchend.

        """

        logger.info('creating frequency breakdown')
        breakdown = self.corpus.counts.dump(
            df_dump=self.df,
            start='match', end='matchend',
            p_atts=p_atts, strategy=1
        )

        breakdown = fold_df(breakdown, flags)

        return breakdown

    def matches(self):
        """
        :return: cpos of (match .. matchend) regions
        :rtype: set
        """
        if self._matches is None:
            f1 = set()
            for match, matchend in self.df.index:
                f1.update(range(match, matchend + 1))
            self._matches = f1

        return self._matches

    def context(self):
        """
        :return: cpos of (context .. contextend) regions including matches
        :rtype: DataFrame
        """
        if self._context is None:
            self._context = DataFrame.from_records(merge_intervals(
                self.df[['context', 'contextend']].values.tolist()
            ), columns=['context', 'contextend'])

        return self._context

    def marginals(self, start='match', end='matchend', p_atts=['word']):

        return self.corpus.counts.dump(
            self.df, start=start, end=end, p_atts=p_atts, split=True
        )

    def concordance(self, form='simple', p_show=['word'], s_show=[],
                    order='first', cut_off=100, matches=None,
                    slots=None, cwb_ids=False):

        conc = Concordance(
            self.corpus.copy(),
            df_dump=self.df
        )

        return conc.lines(
            form=form,
            p_show=p_show,
            s_show=s_show,
            order=order,
            cut_off=cut_off,
            matches=matches,
            slots=slots,
            cwb_ids=cwb_ids
        )

    def collocates(self, p_query=['lemma'], mws=20, window=5, order='O11',
                   cut_off=100, ams=None, min_freq=2,
                   frequencies=True, flags=None, marginals='corpus',
                   show_negative=False):

        mws = max(mws, window)

        coll = Collocates(
            self.corpus.copy(),
            df_dump=self.df,
            p_query=p_query,
            mws=mws
        )

        return coll.show(
            window=window,
            order=order,
            cut_off=cut_off,
            ams=ams,
            min_freq=min_freq,
            frequencies=frequencies,
            flags=flags,
            marginals=marginals,
            show_negative=show_negative
        )

    def keywords(self, p_query=['lemma'], order='O11', cut_off=100,
                 ams=None, min_freq=2, frequencies=True, flags=None):

        kw = Keywords(
            self.corpus.copy(),
            self.df,
            p_query
        )

        return kw.show(
            order=order,
            cut_off=cut_off,
            ams=ams,
            min_freq=min_freq,
            frequencies=frequencies,
            flags=flags
        )


class Dumps:
    """ collection of dumps, i.e. query matches in different subcorpora """

    def __init__(self, corpus, s_dict, s_att='text_id'):
        """
        :param Corpus corpus: corpus to work on
        :param dict s_dict: dicitonary of {subcorpus_name: set of values for s_att}
        :param str s_att: s-attribute that is used for definition of subcorpora
        """

        self.corpus = corpus.copy()
        self.s_dict = s_dict
        self.s_att = s_att

        logger.info("creating subcorpora ...")
        df_spans = corpus.dump_from_s_att(s_att)
        dumps = dict()
        for i, s in enumerate(self.s_dict.keys()):
            logger.info("... subcorpus %d of %d" % (i + 1, len(s_dict)))
            df_dump = df_spans.loc[df_spans[s_att].isin(s_dict[s])]
            dumps[s] = Dump(corpus.copy(), df_dump, name_cqp=None)

        self.dumps = dumps

    def keywords(self, p_query=['lemma'], order='log_likelihood', cut_off=100,
                 ams=None, min_freq=2, frequencies=True, flags=None,
                 subset=None):
        """
        :return: dictionary of {subcorpus_name: table}
        :rtype: dict
        """

        subset = list(self.s_dict.keys()) if subset is None else subset

        logger.info("computing keyword tables ...")
        # TODO: multiproc
        tables = dict()
        i = 0
        for s in subset:
            i += 1
            logger.info("... table %d of %d" % (i, len(subset)))
            dump = self.dumps[s]
            tables[s] = dump.keywords(
                p_query=p_query, order=order, cut_off=cut_off,
                ams=ams, min_freq=min_freq,
                frequencies=frequencies, flags=flags
            )

        return tables

    def collocates(self, cqp_query, window=5, p_query=['lemma'],
                   order='log_likelihood', cut_off=100, ams=None, min_freq=2,
                   frequencies=True, flags=None, subset=None, context_break=None,
                   reference='local'):
        """
        reference:
        .. local: window freq. compared to marginals of subcorpus (excl. nodes)
        .. global: window freq. compared to marginals in whole corpus (excl. nodes)
        .. DataFrame:

        :param str reference: 'local' | 'global' | DataFrame
        :return: dictionary of {subcorpus_name: table}
        :rtype: dict
        """

        subset = list(self.s_dict.keys()) if subset is None else subset
        context_break = self.s_att if context_break is None else context_break

        # run query once and extend dump
        dump_glob = self.corpus.query(
            cqp_query,
            context=window,
            context_break=context_break
        ).df
        df_glob = self.corpus.dump2satt(dump_glob, self.s_att)

        logger.info("computing collocate tables ...")
        # TODO: multiproc
        tables = dict()
        i = 0
        for s in subset:
            i += 1
            logger.info("... table %d of %d" % (i, len(subset)))

            # determine reference frequencies
            if isinstance(reference, str):
                if reference == 'local':
                    # get local marginals
                    marginals = self.corpus.counts.dump(
                        self.dumps[s].df, split=True, p_atts=p_query
                    )
                elif reference == 'global':
                    marginals = 'corpus'

            # create collocates table
            df_loc = df_glob.loc[
                df_glob[self.s_att].isin(self.s_dict[s])
            ]
            collocates = Collocates(
                self.corpus, df_loc, p_query=p_query, mws=window
            )
            tables[s] = collocates.show(
                window=window, order=order, cut_off=cut_off,
                ams=ams, min_freq=min_freq, frequencies=frequencies,
                flags=flags, marginals=marginals
            )

        return tables



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import pkgutil
import re
from functools import wraps
from timeit import default_timer

# requirements
import numpy as np
from pandas import NA
from unidecode import unidecode

logger = logging.getLogger(__name__)


def time_it(func):
    """
    decorator for printing the execution time of a function call
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        start = default_timer()
        result = func(*args, **kwargs)
        end = default_timer()
        name = func.__name__
        logger.info("%s took %s seconds" % (name, round(end - start, 2)))
        return result
    return wrapper


###########
# ANCHORS #
###########
def chunk_anchors(lst, n, exclude={0, 1}):
    """yields n-sized chunks from lst without excluded values"""
    lst_rm = [i for i in lst if i not in exclude]
    for i in range(0, len(lst_rm), n):
        yield lst_rm[i:i + n]


def apply_correction(row, correction):
    """Applies correction to one row in column.

    :param list row: row of dump with [value, context, contextend]
    :param int correction: offset to add/subtract from position
    """

    value, lower_bound, upper_bound = row

    if value == -1:
        # missing value
        return value

    value += correction
    if value < lower_bound:
        value = lower_bound
    if value > upper_bound:
        value = upper_bound
    return value


def correct_anchors(df, corrections):
    """Corrects df columns via dictionary of corrections

    :param DataFrame df: dump
    :param dict corrections: {anchor_name (0-9, column in df): offset (int)}
    """
    for correction in corrections:
        if correction in df.columns:
            logger.info('correcting anchor %d by offset %d' % (
                correction, corrections[correction]
            ))
            df[correction] = df[
                [correction, 'context', 'contextend']
            ].apply(lambda x: apply_correction(x, corrections[correction]), axis=1)
        else:
            logger.warning('anchor "%s" not in dataframe' % str(correction))
    return df


###########
# QUERIES #
###########
def cqp_escape(token):
    """ escape CQP meta-characters

    :param str token: string to escape
    :return: escaped string
    :rtype: str

    """
    return token.translate(str.maketrans({
        ".": r"\.",
        "?": r"\?",
        "*": r"\*",
        "+": r"\+",
        "|": r"\|",
        "(": r"\(",
        ")": r"\)",
        "[": r"\[",
        "]": r"\]",
        "{": r"\{",
        "}": r"\}",
        "^": r"\^",
        "$": r"\$",
        '"': r"\"",
        "'": r"\'"
    }))


def format_cqp_query(items, p_query='word', s_query=None,
                     flags="", escape=True):
    """ wrapper for easy queries

    :return: valid cqp query
    :rtype: str
    """

    # gather all MWUs
    mwu_queries = list()
    single_items = list()
    for item in items:

        # escape item (optional)
        if escape:
            item = cqp_escape(item)

        # split items on white-space
        tokens = item.split(" ")
        if len(tokens) > 1:
            mwu_query = ""

            # convert to CQP syntax considering p-att and flags
            for token in tokens:
                mwu_query += '[{p_query}="{token}"{flags}]'.format(
                    p_query=p_query, token=token, flags=flags
                )

            # add MWU item to list
            mwu_queries.append("(" + mwu_query + ")")
        else:
            single_items.append(tokens[0])

    singles_query = "([" + p_query + "=" + '"' + "|".join(single_items) + '"' + flags + "])"
    queries = mwu_queries + [singles_query]
    # disjunctive join
    query = ' | '.join(queries)

    # add s_query (optional)
    if s_query is not None:
        query = '({query}) within {s_query};'.format(
            query=query, s_query=s_query
        )

    # return
    return query


def find_all(pattern, string):
    p = re.compile(pattern)
    matches = dict()
    for m in p.finditer(string):
        matches[m.group()] = [m.start(), m.end()]
    return matches


def preprocess_query(query):
    """
    - get s_query and strip 'within' statement
    - get anchors
    - get macros
    - get wordlists

    :return: raw query, s_query, anchors, macros, wordlists
    :rtype: dict
    """

    # get s_query and strip 'within' statement
    s_query = None
    query = query.rstrip(";")
    tokens = query.split(" ")
    if len(tokens) > 2:
        if tokens[-2] == "within":
            query = " ".join(tokens[:-2])
            s_query = tokens[-1]

    # find anchors, macros, wordlists
    anchors = [int(a[1]) for a in find_all(r"@\d", query).keys()]
    macros = list(find_all(r"/[A-Za-z_]+\[\d*\]", query).keys())
    wordlists = list(find_all(r"\$[A-Za-z_]+", query).keys())

    return {
        'query': query,
        's_query': s_query,
        'anchors': anchors,
        'macros': macros,
        'wordlists': wordlists
    }


#################################
# working on nodes and contexts #
#################################
def _node2cotext(match, matchend, context, contextend):
    """converts the four cpos-values of nodes into a dict of lists of
    cpos, match, and offset

    """
    # get lists
    cpos_list = list(range(context, contextend + 1))
    match_list = [match] * len(cpos_list)
    offset_list = [
        (cpos - match) if cpos < match else
        (cpos - matchend) if cpos > matchend else
        0 for cpos in cpos_list
    ]

    # return object
    result = {
        'match_list': match_list,
        'cpos_list': cpos_list,
        'offset_list': offset_list
    }

    return result


node2cotext = np.vectorize(_node2cotext)


def merge_intervals(inter, start_index=0):
    """ for merging contexts """
    for i in range(start_index, len(inter)-1):
        if inter[i][1] >= inter[i+1][0]:
            new_start = inter[i][0]
            new_end = inter[i+1][1]
            inter[i] = [new_start, new_end]
            del inter[i+1]
            return merge_intervals(inter.copy(), start_index=i)
    return inter


def calculate_offset(row):
    """ calculates offset of y to x """

    # necessary values
    match_x = row['match_x']
    match_y = row['match_y']

    # if match_y is missing, offset is missing
    if match_y is NA or match_y == -1:
        return NA

    # init matchends if necessary
    if 'matchend_x' in row.keys():
        matchend_x = row['matchend_x']
    else:
        matchend_x = match_x
    if 'matchend_y' in row.keys():
        matchend_y = row['matchend_y']
    else:
        matchend_y = match_y

    # y ... x
    if match_x > matchend_y:
        offset = matchend_y - match_x
    # x ... y
    elif matchend_x < match_y:
        offset = match_y - matchend_x
    # overlap
    else:
        offset = 0

    return offset


##################
# s-att handling #
##################
def merge_s_atts(s_query, s_break, s_meta):
    """ DEPRECATED consistencize s-atts """
    # s_query < s_break < s_meta

    # case 1.1: only s_query
    if s_break is None and s_meta is None:
        s_break = s_meta = s_query

    # case 1.2: only s_break
    elif s_query is None and s_meta is None:
        s_query = s_meta = s_break

    # case 1.3: only s_meta
    elif s_query is None and s_break is None:
        s_query = s_break = s_query

    # case 2.1: s_query and s_break
    elif s_meta is None:
        s_meta = s_break

    # case 2.2: s_query and s_meta
    elif s_break is None:
        s_break = s_meta

    # useless cases
    # case 2.3: s_break and s_meta
    # case 3: all given

    if s_meta is not None and not s_meta.endswith("_id"):
        s_meta = s_meta + "_id"

    logger.info("s_query: %s - s_break: %s - s_meta: %s" % (
        str(s_query), str(s_break), str(s_meta)
    ))
    return s_query, s_break, s_meta


##################
# p-att handling #
##################
def fold_item(item, flags="%cd"):

    if flags is None:
        return item

    # is the item a string or a tuple?
    isstr = isinstance(item, str)
    item = (item, ) if isstr else item

    if "c" in flags:
        # lower-case
        item = [i.lower() for i in item]

    if "d" in flags:
        # TODO align with CWB
        # remove diacritica
        item = [unidecode(i) for i in item]

    item = item[0] if isstr else tuple(item)

    return item


def fold_df(df, flags="%cd"):

    if flags is None:
        return df

    df.index = df.index.map(lambda x: fold_item(x, flags))
    grouped = df.groupby(df.index)
    df = grouped.aggregate(np.sum)
    return df


def filter_df(df, path):
    data = pkgutil.get_data(__name__, path)
    items = set(data.decode().split("\n"))
    return df.loc[~df.index.isin(items)]



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import os
import shlex
import subprocess

from setuptools import Extension, setup

# try:
#     from Cython.Build import cythonize
#     USE_CYTHON = True
# except ImportError:
USE_CYTHON = False              # use cython -2 ccc/cl.pyx instead


########################
# read local resources #
########################
here = os.path.abspath(os.path.dirname(__file__))

# description
with open(os.path.join(here, 'README.md'), mode='rt', encoding='utf-8') as f:
    long_description = f.read()

# version
version = {}
with open(os.path.join(here, 'ccc', 'version.py'), mode='rt', encoding='utf-8') as f:
    exec(f.read(), version)


##############
# cwb-config #
##############
cwb_version = subprocess.run(shlex.split("cwb-config -v"), capture_output=True).stdout.decode().strip()
cwb_bindir = subprocess.run(shlex.split("cwb-config --bindir"), capture_output=True).stdout.decode().strip()
# - effective registry directory or directories:
cwb_registry = subprocess.run(shlex.split("cwb-config -r"), capture_output=True).stdout.decode().strip()
# - libdir (CL library)
cwb_libdir = subprocess.run(shlex.split("cwb-config --libdir"), capture_output=True).stdout.decode().strip()
# - incdir (C header)
cwb_incdir = subprocess.run(shlex.split("cwb-config --incdir"), capture_output=True).stdout.decode().strip()
# - compiler flags for linking against CL library
cwb_compiler_flags = subprocess.run(shlex.split("cwb-config -I"), capture_output=True).stdout.decode().strip()
#  -linker flags for linking against CL library
cwb_linker_flags = subprocess.run(shlex.split("cwb-config -L"), capture_output=True).stdout.decode().strip()


####################################
# define (and compile) C-extension #
####################################
libraries = [t[2:] for t in shlex.split(cwb_linker_flags) if t.startswith("-l")]
ccc_cl = Extension(
    name="ccc.cl",
    sources=['ccc/cl' + ('.pyx' if USE_CYTHON else '.c')],
    include_dirs=[cwb_incdir],  # list of directories to search for C/C++ header files
    library_dirs=[cwb_libdir],  # list of directories to search for C/C++ libraries at link time
    libraries=libraries         # list of library names (not filenames or paths) to link against
)

# cythonize?
extensions = [ccc_cl]
if USE_CYTHON:
    from Cython.Build import cythonize
    extensions = cythonize(extensions)


#################
# actual set-up #
#################
setup(
    name="cwb-ccc",
    version=version["__version__"],
    description="CWB wrapper to extract concordances and score frequency lists",
    license='GNU General Public License v3 or later (GPLv3+)',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Philipp Heinrich",
    author_email="philipp.heinrich@fau.de",
    url="https://github.com/ausgerechnet/cwb-ccc",
    packages=[
        'ccc'
    ],
    ext_modules=extensions,
    python_requires='>=3.6.2',
    install_requires=[
        "wheel>=0.37.1",
        "association-measures>=0.2.4",
        "pandas>=1.1.5",
        "numexpr>=2.7.1",
        "Bottleneck>=1.3.4",
        "unidecode>=1.3.4",
        "pyyaml>=6.0"
    ],
    classifiers=[
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Development Status :: 4 - Beta",
        "Operating System :: Unix",
        "Programming Language :: Python :: 3",
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Cython',
    ],
)
